/*
 * Copyright (c) 2003 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bbs_thread_ui.c,v 1.16 2003/07/13 14:53:51 fuyu Exp $
 */

#include "config.h"

#include "ochusha_private.h"
#include "ochusha.h"
#include "ochusha_thread_2ch.h"

#include "ochusha_ui.h"
#include "bbs_thread_ui.h"
#include "bbs_thread_view.h"
#include "boardlist_ui.h"
#include "icon_label.h"

#include "htmlutils.h"
#include "worker.h"
#include "utils.h"

#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define DEBUG_POPUP	0

typedef struct _PopupContext PopupContext;

static void link_popup_menu_callback(gpointer data, guint action,
				     GtkWidget *widget);
static void link_popup_menu_destructed_cb(gpointer data);
static gboolean link_mouse_over_cb(BBSThreadView *view, GdkEventMotion *event,
				   OchushaBBSThread *thread, const gchar *link,
				   OchushaApplication *application);
static gboolean link_mouse_out_cb(BBSThreadView *view, GdkEventAny *event,
				  OchushaBBSThread *thread, const gchar *link,
				  OchushaApplication *application);
static gboolean link_mouse_press_cb(BBSThreadView *view, GdkEventButton *event,
				    OchushaBBSThread *thread,
				    const gchar *link,
				    OchushaApplication *application);
static gboolean link_mouse_release_cb(BBSThreadView *view,
				      GdkEventButton *event,
				      OchushaBBSThread *thread,
				      const gchar *link,
				      OchushaApplication *application);

static void show_popup(OchushaApplication *application,
		       OchushaBBSThread *thread,
		       const gchar *link, gint x_pos, gint y_pos,
		       BBSThreadView *parent);
static gboolean popup_delay_timeout(PopupContext *popup_context);
static void show_popup_real(PopupContext *popup_context);
static void hide_popup(PopupContext *popup_context);

static void render_bbs_thread(WorkerThread *employee, gpointer args);
static gboolean start_thread_rendering(OchushaBBSThread *thread,
				       const gchar *title,
				       gpointer user_data);
static gboolean render_response(OchushaBBSThread *thread, gint number,
				const OchushaBBSResponse *response,
				gpointer user_data);
static gboolean render_broken_response(OchushaBBSThread *thread, gint number,
				       gpointer user_data);
static gboolean end_thread_rendering(OchushaBBSThread *thread,
				     gboolean finished,
				     gpointer user_data);
static void start_element_cb(void *context, const char *name,
			     const char *const attrs[]);
static void end_element_cb(void *context, const char *name);
static void characters_cb(void *context, const char *ch, int len);

static pthread_mutex_t bbs_thread_ui_lock;

/* ơС˥ᥤȤʤɤɽ뤿context ID */
static guint link_field_id;

/* Ǹ˥ơСɽåФmessage ID */
static guint message_id;

#define BBS_THREAD_UI_LOCK				\
  if (pthread_mutex_lock(&bbs_thread_ui_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't lock a mutex.\n");	\
      abort();						\
    }

#define BBS_THREAD_UI_UNLOCK				\
  if (pthread_mutex_unlock(&bbs_thread_ui_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't unlock a mutex.\n");	\
      abort();						\
    }


typedef struct _RendererContext
{
  ElementHandler handler;
  BBSThreadView *view;
  OchushaBBSThread *thread;
  OchushaApplication *application;
  gint last_read_number;

  gchar *link;
  gchar *link_text_buffer;
  guint link_text_buffer_size;
  guint link_text_length;

  GHashTable *tag_table;

  GtkTextTag *thread_title_tag;
  GtkTextTag *response_name_tag;
  GtkTextTag *response_paragraph_tag;
  GtkTextTag *bold_tag;
} RendererContext;


#define DEFAULT_LINK_BUFFER_SIZE	512
#define DEFAULT_BUFFER_SIZE		4096

static GHashTable *tag_table = NULL;
static int stack_depth = 0;


struct _PopupContext
{
  RendererContext renderer_context;

  GtkWidget *popup_window;
  GtkContainer *popup_frame;

  PopupContext *parent_context;

  gchar *url;
  guint delay_id;
  guint close_delay_id;
  gint x_pos;
  gint y_pos;

  gboolean now_pointed;
};


/* PopupϢ */
#if !ENABLE_MULTIPLE_POPUPS
gchar popup_renderer_default_text_buffer[DEFAULT_BUFFER_SIZE];
PopupContext global_popup_context =
{
  {
    {
      0,
      start_element_cb,
      end_element_cb,
      NULL,
      characters_cb
    },
    NULL,
    NULL,
    NULL,
    -1,	/* ְ㤨ƥޡդʤ褦 */
  
    NULL,
    popup_renderer_default_text_buffer,
    DEFAULT_BUFFER_SIZE,
    0,

    NULL,
    NULL,
    NULL,
    NULL,
  },

  NULL,		/* popup_window */
  NULL,		/* popup_frame */

  NULL,		/* parent_context */

  NULL,		/* url */
  0,		/* delay_id */
  0,		/* close_delay_id */
  0,		/* x_pos */
  0,		/* y_pos */

  FALSE,	/* now_pointed */
};
#endif


enum {
  LINK_POPUP_MENU_OPEN_LINK = 1,
  LINK_POPUP_MENU_OPEN_LINK_IN_TAB,
  LINK_POPUP_MENU_OPEN_LINK_WITH_WEB_BROWSER,
  LINK_POPUP_MENU_SAVE_LINK_AS,
  LINK_POPUP_MENU_COPY_LINK_LOCATION,
};


static GtkItemFactory *link_popup_menu_item_factory = NULL;
static GQuark bbs_thread_gui_info_id;
static GQuark popup_context_id;
static GQuark popup_parent_context_id;


void
prepare_thread_ui_initialization(OchushaApplication *application)
{
  if (pthread_mutex_init(&bbs_thread_ui_lock, NULL) != 0)
    {
      fprintf(stderr, "Couldn't init a mutex.\n");
      abort();
    }

  bbs_thread_gui_info_id = g_quark_from_static_string("BBSThreadUI::GUIInfo");
  popup_context_id = g_quark_from_static_string("BBSThreadUI::PopupContext");
  popup_parent_context_id
    = g_quark_from_static_string("BBSThreadUI::PopupParentContext");
}


static GtkTextTag *thread_title_tag;
static GtkTextTag *response_name_tag;
static GtkTextTag *response_paragraph_tag;
static GtkTextTag *bold_tag;


static void
initialize_popup_context(PopupContext *popup_context,
			 OchushaApplication *application)
{
  GtkWidget *frame;
  GtkWindow *parent_window;

#if DEBUG_POPUP
  fprintf(stderr, "initialize_popup_context(%p, %p)\n",
	  popup_context, application);
#endif

  popup_context->popup_window = gtk_window_new(GTK_WINDOW_POPUP);
  gtk_window_set_resizable(GTK_WINDOW(popup_context->popup_window),
			   FALSE);

  frame = gtk_frame_new(NULL);
  gtk_widget_show(frame);
  popup_context->popup_frame = GTK_CONTAINER(frame);
  gtk_container_add(GTK_CONTAINER(popup_context->popup_window), frame);
  
  popup_context->renderer_context.application = application;

  popup_context->renderer_context.thread_title_tag = thread_title_tag;
  popup_context->renderer_context.response_name_tag = response_name_tag;
  popup_context->renderer_context.response_paragraph_tag
    = response_paragraph_tag;
  popup_context->renderer_context.bold_tag = bold_tag;
  popup_context->renderer_context.tag_table = tag_table;

  /* popup_context->now_pointed = FALSE; */
  parent_window
    = (popup_context->parent_context == NULL
       ? popup_context->renderer_context.application->top_level
       : GTK_WINDOW(popup_context->parent_context->popup_window));
  gtk_window_set_transient_for(GTK_WINDOW(popup_context->popup_window),
			       parent_window);
}


void
initialize_thread_ui(OchushaApplication *application)
{
  GtkWidget *dummy_widget;
  BBSThreadView *dummy_view;

  /* ݥåץåץ˥塼ν */
  GtkItemFactoryEntry menu_items[] =
    {
      {
	_("/_Open Link"),				/* menu path */
	NULL,						/* accelerator */
	link_popup_menu_callback,			/* callback_func */
	LINK_POPUP_MENU_OPEN_LINK,			/* act */
	NULL						/* type */
      },
      {
	_("/Open Link in New _Tab"),			/* menu path */
	NULL,						/* accelerator */
	link_popup_menu_callback,			/* callback_func */
	LINK_POPUP_MENU_OPEN_LINK_IN_TAB,		/* act */
	NULL						/* type */
      },
      {
	_("/Open Link with _Web Browser"),		/* menu path */
	NULL,						/* accelerator */
	link_popup_menu_callback,			/* callback_func */
	LINK_POPUP_MENU_OPEN_LINK_WITH_WEB_BROWSER,	/* act */
	NULL						/* type */
      },
      {
	_("/-------"),
	NULL,
	NULL,
	0,
	"<Separator>"
      },
      {
	_("/_Save Link As..."),				/* menu path */
	NULL,						/* accelerator */
	link_popup_menu_callback,			/* callback_func */
	LINK_POPUP_MENU_SAVE_LINK_AS,			/* act */
	NULL						/* type */
      },
      {
	_("/_Copy Link Location"),			/* menu path */
	NULL,						/* accelerator */
	link_popup_menu_callback,			/* callback_func */
	LINK_POPUP_MENU_COPY_LINK_LOCATION,		/* act */
	NULL						/* type */
      },
    };
  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);
  link_popup_menu_item_factory = gtk_item_factory_new(GTK_TYPE_MENU,
						      "<ThreadLinkPopup>",
						      NULL);
  gtk_item_factory_create_items(link_popup_menu_item_factory,
				nmenu_items, menu_items, application);

  /* ̤εǽ˥塼Ĥ֤ */
  dummy_widget = gtk_item_factory_get_widget_by_action(
					link_popup_menu_item_factory,
					LINK_POPUP_MENU_SAVE_LINK_AS);
  if (dummy_widget != NULL)
    gtk_widget_set_sensitive(dummy_widget, FALSE);

  g_return_if_fail(tag_table == NULL);

  link_field_id = gtk_statusbar_get_context_id(application->statusbar,
					       "response-link-field");

  dummy_widget = bbs_thread_view_new(NULL);
  dummy_view = BBS_THREAD_VIEW(dummy_widget);

  thread_title_tag
    = bbs_thread_view_get_tag_by_name(dummy_view, "thread_title");
  response_name_tag
    = bbs_thread_view_get_tag_by_name(dummy_view, "response_name");
  response_paragraph_tag
    = bbs_thread_view_get_tag_by_name(dummy_view, "response_paragraph");
  bold_tag = bbs_thread_view_get_tag_by_name(dummy_view, "bold");

  /* Τ */
  gtk_object_sink(GTK_OBJECT(dummy_view));

  {
#if 0
    GtkTextTag *tag;
    GtkTextTagTable *default_tag_table = bbs_thread_view_class_get_default_tag_table(BBS_THREAD_VIEW_GET_CLASS(job_args->view));
#endif
    tag_table = g_hash_table_new(g_str_hash, g_str_equal);

    /* TODO: äȡ */
    g_hash_table_insert(tag_table, "b", bold_tag);
  }

#if !ENABLE_MULTIPLE_POPUPS
  initialize_popup_context(&global_popup_context, application);
#endif  
}


static void
bbs_thread_info_free(BBSThreadGUIInfo *info)
{
  if (info->offsets != NULL)
    {
      G_FREE(info->offsets);
      info->offsets = NULL;
    }

  G_FREE(info);
}


BBSThreadGUIInfo *
ensure_bbs_thread_info(OchushaBBSThread *thread)
{
  BBSThreadGUIInfo *info;
  BBS_THREAD_UI_LOCK
    {
      info = g_object_get_qdata(G_OBJECT(thread), bbs_thread_gui_info_id);
      if (info == NULL)
	{
	  info = G_NEW0(BBSThreadGUIInfo, 1);
	  g_object_set_qdata_full(G_OBJECT(thread), bbs_thread_gui_info_id,
				  info, (GDestroyNotify)bbs_thread_info_free);
	}
    }
  BBS_THREAD_UI_UNLOCK;
  return info;
}


typedef struct _BBSThreadJobArgs
{
  OchushaApplication *application;
  OchushaBBSThread *thread;
  BBSThreadView *view;
  IconLabel *tab_label;
  OchushaAsyncBuffer *buffer;
  gint start;
  gint number_of_responses;
  gboolean refresh;
} BBSThreadJobArgs;


GtkWidget *
open_bbs_thread(OchushaApplication *application, OchushaBBSThread *thread,
		IconLabel *tab_label)
{
  GtkWidget *thread_view;
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(thread);

  if (info == NULL)
    {
      fprintf(stderr, "Out of memory.\n");
      return NULL;
    }

  BBS_THREAD_UI_LOCK
    {
      WorkerJob *job;
      BBSThreadJobArgs *job_args;
      OchushaAsyncBuffer *buffer;

      thread_view = bbs_thread_view_new(thread);
      gtk_widget_show(thread_view);

      job = G_NEW0(WorkerJob, 1);
      job_args = G_NEW0(BBSThreadJobArgs, 1);
      buffer = ochusha_bbs_thread_get_responses_source(thread,
				application->broker,
				OCHUSHA_NETWORK_BROKER_CACHE_TRY_UPDATE);

      if (buffer == NULL)
	{
	  fprintf(stderr, "Couldn't access network!\n");
	  G_FREE(job_args);
	  G_FREE(job);
	  gtk_object_sink(GTK_OBJECT(thread_view));

	  BBS_THREAD_UI_UNLOCK;
	  return NULL;
	}

      setup_for_tab_label_animation(buffer, tab_label);

      info->response_source_buffer = buffer;
      info->rendering_done = FALSE;

      job_args->application = application;
      job_args->thread = thread;
      job_args->view = BBS_THREAD_VIEW(thread_view);
      job_args->buffer = buffer;
      job_args->start = 0;
      job_args->number_of_responses = -1;
      job_args->refresh = FALSE;
      job_args->tab_label = tab_label;

      job->canceled = FALSE;
      job->job = render_bbs_thread;
      job->args = job_args;

      g_signal_connect(G_OBJECT(thread_view), "link_mouse_over",
		       G_CALLBACK(link_mouse_over_cb), application);
      g_signal_connect(G_OBJECT(thread_view), "link_mouse_out",
		       G_CALLBACK(link_mouse_out_cb), application);
      g_signal_connect(G_OBJECT(thread_view), "link_mouse_press",
		       G_CALLBACK(link_mouse_press_cb), application);
      g_signal_connect(G_OBJECT(thread_view), "link_mouse_release",
		       G_CALLBACK(link_mouse_release_cb), application);

#if DEBUG_ASYNC_BUFFER_MOST
      fprintf(stderr, "open_bbs_thread: ref AsyncBuffer(%p)\n", buffer);
      fprintf(stderr, "* before ref: buffer->ref_count=%d\n",
	      G_OBJECT(buffer)->ref_count);
#endif
      /* öref_count䤹 */
      g_object_ref(G_OBJECT(buffer));
      g_object_ref(G_OBJECT(tab_label));
      g_object_ref(G_OBJECT(thread_view));
      commit_job(job);
    }
  BBS_THREAD_UI_UNLOCK;

  return thread_view;
}


static void
link_popup_menu_callback(gpointer data, guint action, GtkWidget *widget)
{
  OchushaApplication *application = data;
  const char *url = gtk_item_factory_popup_data(link_popup_menu_item_factory);
#if DEBUG_GUI_MOST
  fprintf(stderr, "link_popup_menu_callback: application=%p, url=\"%s\"(@%p), action=%d\n",
	  application, url, url, action);
#endif
  switch (action)
    {
    case LINK_POPUP_MENU_OPEN_LINK:
      ochusha_open_url(application, url, FALSE, FALSE);
      break;

    case LINK_POPUP_MENU_OPEN_LINK_IN_TAB:
      ochusha_open_url(application, url, TRUE, FALSE);
      break;
      
    case LINK_POPUP_MENU_OPEN_LINK_WITH_WEB_BROWSER:
      ochusha_open_url(application, url, FALSE, TRUE);
      break;

    case LINK_POPUP_MENU_COPY_LINK_LOCATION:
      ochusha_clipboard_set_text(application, url);
      break;

    case LINK_POPUP_MENU_SAVE_LINK_AS:
      fprintf(stderr, "Not implemented yet.\n");
      break;
    }
}


static void
link_popup_menu_destructed_cb(gpointer data)
{
#if DEBUG_GUI_MOST
  fprintf(stderr, "link_popup_menu_destructed_cb: url=\"%s\"\n", (char *)data);
#endif
  G_FREE(data);
}


static gboolean
link_mouse_over_cb(BBSThreadView *view, GdkEventMotion *event,
		   OchushaBBSThread *thread, const gchar *link,
		   OchushaApplication *application)
{
  GtkStatusbar *statusbar = application->statusbar;
#if DEBUG_GUI_MOST
  fprintf(stderr, "link_mouse_over_cb: link=\"%s\"\n", link);
#endif
  if (message_id != 0)
    {
      gtk_statusbar_remove(statusbar, link_field_id, message_id);
      stack_depth--;
    }

#if DEBUG_POPUP
  fprintf(stderr, "link_mouse_over_cb: show_popup()\n");
#endif

  if (application->enable_popup_response)
    show_popup(application, thread, link, event->x_root, event->y_root, view);

  stack_depth++;
  if (stack_depth == 1)
    message_id = gtk_statusbar_push(statusbar, link_field_id, link);

  return FALSE;
}


static void
popup_context_free(PopupContext *popup_context)
{
#if DEBUG_POPUP
  fprintf(stderr, "popup_context_free(%p)\n", popup_context);
#endif

  if (popup_context->renderer_context.handler.converter != 0)
    {
      iconv_close(popup_context->renderer_context.handler.converter);
      popup_context->renderer_context.handler.converter = 0;
    }

  if (popup_context->renderer_context.link_text_buffer != NULL)
    {
      G_FREE(popup_context->renderer_context.link_text_buffer);
      popup_context->renderer_context.link_text_buffer = NULL;
    }

  if (popup_context->url != NULL)
    {
      G_FREE(popup_context->url);
      popup_context->url = NULL;
    }

  if (popup_context->delay_id != 0)
    {
      g_source_remove(popup_context->delay_id);
      popup_context->delay_id = 0;
    }

  if (popup_context->close_delay_id != 0)
    {
      g_source_remove(popup_context->close_delay_id);
      popup_context->close_delay_id = 0;
    }

  if (popup_context->popup_window != NULL)
    {
      gtk_widget_destroy(popup_context->popup_window);
      popup_context->popup_window = NULL;
    }

  G_FREE(popup_context);
}


/*
 * MEMO: gdk_threads_enter()ĶƤӽФȡ
 */
static PopupContext *
ensure_popup_context(BBSThreadView *parent, OchushaApplication *application)
{
#if ENABLE_MULTIPLE_POPUPS
  RendererContext *renderer_context;
  PopupContext *popup_context = g_object_get_qdata(G_OBJECT(parent),
						   popup_context_id);
  if (popup_context != NULL)
    return popup_context;

  popup_context = G_NEW0(PopupContext, 1);

  renderer_context = &popup_context->renderer_context;

  /* renderer_context->handler.converter = 0; */
  renderer_context->handler.startElement = start_element_cb;
  renderer_context->handler.endElement = end_element_cb;
  /* renderer_context->handler.entityReference = NULL; */
  renderer_context->handler.characters = characters_cb;

  /* renderer_context->view = NULL; */
  /* renderer_context->thread = NULL; */
  renderer_context->application = application;
  renderer_context->last_read_number = -1;

  /* renderer_context->link = NULL; */
  renderer_context->link_text_buffer = G_NEW0(gchar, DEFAULT_LINK_BUFFER_SIZE);
  renderer_context->link_text_buffer_size = DEFAULT_LINK_BUFFER_SIZE;
  /* renderer_context->link_text_length = 0; */

  popup_context->parent_context = g_object_get_qdata(G_OBJECT(parent),
						     popup_parent_context_id);
  initialize_popup_context(popup_context, application);
  g_object_set_qdata_full(G_OBJECT(parent), popup_context_id, popup_context,
			  (GDestroyNotify)popup_context_free);


  return popup_context;
#else
  return &global_popup_context;
#endif
}


static gboolean
link_mouse_out_cb(BBSThreadView *view, GdkEventAny *event,
		  OchushaBBSThread *thread, const gchar *link,
		  OchushaApplication *application)
{
  PopupContext *popup_context;
  GtkStatusbar *statusbar = application->statusbar;

  popup_context = ensure_popup_context(view, application);

  if (message_id != 0)
    {
      gtk_statusbar_remove(statusbar, link_field_id, message_id);
      message_id = 0;
    }
  stack_depth--;

#if DEBUG_POPUP
  fprintf(stderr, "link_mouse_out_cb: hide_popup(%p)\n", popup_context);
#endif
  hide_popup(popup_context);

  return FALSE;
}


static gboolean
link_mouse_press_cb(BBSThreadView *view, GdkEventButton *event,
		    OchushaBBSThread *thread, const gchar *link,
		    OchushaApplication *application)
{
  if (link == NULL)
    return FALSE;

  if (event->button == 3)
    gtk_item_factory_popup_with_data(link_popup_menu_item_factory,
				     G_STRDUP(link),
				     link_popup_menu_destructed_cb,
				     event->x_root, event->y_root,
				     event->button,
				     gtk_get_current_event_time());
  else
    ochusha_open_url(application, link, event->button == 2, FALSE);

  return TRUE;
}


static gboolean
link_mouse_release_cb(BBSThreadView *view, GdkEventButton *event,
		      OchushaBBSThread *thread, const gchar *link,
		      OchushaApplication *application)
{
  return FALSE;
}


#define DEFAULT_SUBJECT_COLOR	"red"
#define DEFAULT_NAME_COLOR	"green"
#define DEFAULT_LINK_COLOR	"blue"


static void
render_bbs_thread(WorkerThread *employee, gpointer args)
{
  BBSThreadJobArgs *job_args = (BBSThreadJobArgs *)args;
  OchushaApplication *application = job_args->application;
  OchushaBBSThread *thread = job_args->thread;
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(thread);
  OchushaAsyncBuffer *buffer = job_args->buffer;
  gchar default_text_buffer[DEFAULT_BUFFER_SIZE];
  iconv_helper *helper = ochusha_bbs_thread_get_response_iconv_helper(thread);
  RendererContext context =
    {
      {
	helper != NULL
	? iconv_open("UTF-8",
		     ochusha_bbs_thread_get_response_character_encoding(thread))
	: iconv_open("UTF-8//IGNORE",
		     ochusha_bbs_thread_get_response_character_encoding(thread)),
	helper,
	start_element_cb,
	end_element_cb,
	NULL,
	characters_cb
      },
      job_args->view,
      thread,
      application,
      thread->number_of_responses_read,

      NULL,
      default_text_buffer,
      DEFAULT_BUFFER_SIZE,
      0,

      tag_table,

      bbs_thread_view_get_tag_by_name(job_args->view, "thread_title"),
      bbs_thread_view_get_tag_by_name(job_args->view, "response_name"),
      bbs_thread_view_get_tag_by_name(job_args->view, "response_paragraph"),
      bbs_thread_view_get_tag_by_name(job_args->view, "bold"),
  };

  if (info->offsets == NULL)
    info->offsets = G_NEW0(gint, MAX_RESPONSE);
  else 
    memset(info->offsets, 0, MAX_RESPONSE * sizeof(gint));

  if (context.handler.converter == (iconv_t)-1)
    {
      fprintf(stderr, "Out of memory\n");
      BBS_THREAD_UI_LOCK
	{
	  info->rendering_done = TRUE;
	}
      BBS_THREAD_UI_UNLOCK;
      return;
    }

#if DEBUG_GUI_MOST
  {
    gchar *title = convert_string(utf8_to_native, NULL, thread->title, -1);
    fprintf(stderr, "render_bbs_thread: %s\n", title);
    G_FREE(title);
  }
#endif

  if (job_args->start > 0 && !job_args->refresh)
    {
#if DEBUG_GUI
      fprintf(stderr, "XXX: Should this show thread title?");
#endif
    }

  if (job_args->refresh)
    {
      /* ɥ쥹˰*/
      if (info->last_read != NULL)
	bbs_thread_view_delete_mark(job_args->view, info->last_read);
      info->last_read = bbs_thread_view_create_mark(job_args->view);
      info->next_mark = info->last_read;
#if DEBUG_GUI
      fprintf(stderr, "create mark at the tail response.\n");
#endif
    }

  ochusha_bbs_thread_parse_responses(thread, buffer,
				     job_args->start,
				     job_args->number_of_responses,
				     start_thread_rendering,
				     render_response,
				     render_broken_response,
				     end_thread_rendering,
				     &context);

  update_threadlist_entry(job_args->application, thread);

#if DEBUG_WIDGET_MOST
  fprintf(stderr, "render_bbs_thread<done>: view=%p, view->ref_count=%d\n",
	  job_args->view, G_OBJECT(job_args->view)->ref_count);
#endif

  if (job_args->tab_label != NULL)
    g_object_unref(G_OBJECT(job_args->tab_label));

#if DEBUG_WIDGET
  fprintf(stderr, "render_bbs_thread: unref BBSThreadView(%p) ref_count=%d\n",
	  job_args->view, G_OBJECT(job_args->view)->ref_count);
#endif
  g_object_unref(G_OBJECT(job_args->view));

  iconv_close(context.handler.converter);

#if DEBUG_ASYNC_BUFFER_MOST
  fprintf(stderr, "render_bbs_thread: unref AsyncBuffer(%p)\n", buffer);
  fprintf(stderr, "* before unref: buffer->ref_count=%d\n",
	  G_OBJECT(buffer)->ref_count);
#endif
  g_object_unref(G_OBJECT(buffer));

  BBS_THREAD_UI_LOCK
    {
      info->rendering_done = TRUE;
    }
  BBS_THREAD_UI_UNLOCK;

  if (context.link_text_buffer != default_text_buffer)
    G_FREE(context.link_text_buffer);
  G_FREE(args);
}


static gboolean
start_thread_rendering(OchushaBBSThread *thread, const gchar *title,
		       gpointer user_data)
{
  RendererContext *context = (RendererContext *)user_data;
  BBSThreadView *view = context->view;

#if DEBUG_GUI_MOST
  fprintf(stderr, "start_rendering\n");
#endif

  gdk_threads_enter();

  if (context->last_read_number == 0)
    {
      /* ɤॹξ祿ȥߤˤΤǤ˥ޡդ롣*/
      BBSThreadGUIInfo *info = ensure_bbs_thread_info(context->thread);
      info->last_read = bbs_thread_view_create_mark(view);
      info->next_mark = info->last_read;
      bbs_thread_view_scroll_to_mark(view, info->last_read);
    }

  bbs_thread_view_push_tag(view, context->thread_title_tag);

  parse_text(&context->handler, context, title, -1);
  bbs_thread_view_append_text(view, "\n", 1);

  bbs_thread_view_pop_tags(view, context->thread_title_tag);
  gdk_threads_leave();

  return TRUE;	/* go ahead! */
}


#define TEXT_BUFFER_LEN		4096
#define LARGE_COLON		"\357\274\232"

static gboolean
render_response(OchushaBBSThread *thread, gint number,
		const OchushaBBSResponse *response,
		gpointer user_data)
{
  RendererContext *context = (RendererContext *)user_data;
  BBSThreadView *view = context->view;
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(thread);
  char buffer[13];
  const char *cur_pos;
#if DEBUG_GUI_MOST
  fprintf(stderr, "rendering response #%d, last_read_number=%d\n",
	  number, context->last_read_number);
#endif

  if (context->last_read_number >= 0)
    {
      info->offsets[number] = bbs_thread_view_get_current_offset(view);
    }

  if (number == context->last_read_number)
    {
      /* ɥ쥹*/
      info->last_read = bbs_thread_view_create_mark(view);
      info->next_mark = info->last_read;
      gdk_threads_enter();
      bbs_thread_view_scroll_to_mark(view, info->last_read);
      gdk_threads_leave();
    }

  if (thread->number_of_responses_read < number)
    {
      thread->number_of_responses_read = number;
      update_threadlist_entry(context->application, thread);
    }

  snprintf(buffer, 13, "\n%d", number);

  gdk_threads_enter();

#if ENABLE_RENDERING_ON_IDLE
  while (!try_get_rendering_right(context->application,
				  info->response_source_buffer,
				  thread))
    {
      gdk_threads_leave();
      ochusha_async_buffer_wait(info->response_source_buffer,
				"bbs_thread_ui.c: each_response_cb()");
      gdk_threads_enter();
    }
#endif
  bbs_thread_view_append_text(view, buffer, -1);
  bbs_thread_view_append_text(view, LARGE_COLON, 3);

  bbs_thread_view_push_tag(view, context->response_name_tag);
  if (response->mailto == NULL)
    {
      bbs_thread_view_push_tag(view, context->bold_tag);
      parse_text(&context->handler, context, response->name, -1);
      bbs_thread_view_pop_tags(view, context->bold_tag);
    }
  else if (strcmp(response->mailto, "sage") != 0)
    {
      gchar *name = simple_string_canon(response->name, -1,
					context->handler.converter,
					context->handler.helper);
      gchar *mailto = convert_string(context->handler.converter,
				     context->handler.helper,
				     response->mailto, -1);
      bbs_thread_view_append_text_as_link(view, name, -1, mailto);
      G_FREE(mailto);
      G_FREE(name);
    }
  else
    {
      gchar *name = simple_string_canon(response->name, -1,
					context->handler.converter,
					context->handler.helper);
      gchar *mailto = convert_string(context->handler.converter,
				     context->handler.helper,
				     response->mailto, -1);
      bbs_thread_view_append_text_as_sage(view, name, -1, mailto);
      G_FREE(mailto);
      G_FREE(name);
    }
  bbs_thread_view_pop_tags(view, context->response_name_tag);

  bbs_thread_view_append_text(view, LARGE_COLON, 3);
  parse_text(&context->handler, context, response->date_id, -1);

  bbs_thread_view_append_text(view, "\n", 1);


  bbs_thread_view_push_tag(view, context->response_paragraph_tag);
  cur_pos = response->content;
  while (cur_pos != NULL)
    {
      cur_pos = parse_text(&context->handler, context, cur_pos, -1);
      if (cur_pos != NULL)
	{
	  if (*cur_pos == '&')
	    {
	      parse_text(&context->handler, context, "&amp;", 5);
	      cur_pos++;
	    }
	  else if (*cur_pos == '<')
	    {
	      parse_text(&context->handler, context, "&lt;", 4);
	      cur_pos++;
	    }
	  else
	    break;	/*  */
	}
    }

  bbs_thread_view_pop_tags(view, context->response_paragraph_tag);

  bbs_thread_view_append_text(view, "\n", 1);

  gdk_threads_leave();

  return TRUE;
}


static gboolean
render_broken_response(OchushaBBSThread *thread, gint number,
		       gpointer user_data)
{
  RendererContext *context = (RendererContext *)user_data;
  BBSThreadView *view = context->view;
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(thread);
  char buffer[13];
#if DEBUG_GUI_MOST
  fprintf(stderr, "rendering response #%d, last_read_number=%d\n",
	  number, context->last_read_number);
#endif

  if (context->last_read_number >= 0)
    {
      info->offsets[number] = bbs_thread_view_get_current_offset(view);
    }

  if (number == context->last_read_number)
    {
      /* ɥ쥹*/
      info->last_read = bbs_thread_view_create_mark(view);
      info->next_mark = info->last_read;
      gdk_threads_enter();
      bbs_thread_view_scroll_to_mark(view, info->last_read);
      gdk_threads_leave();
    }

  if (thread->number_of_responses_read < number)
    {
      thread->number_of_responses_read = number;
      update_threadlist_entry(context->application, thread);
    }

  snprintf(buffer, 13, "\n%d", number);

  gdk_threads_enter();

#if ENABLE_RENDERING_ON_IDLE
  while (!try_get_rendering_right(context->application,
				  info->response_source_buffer,
				  thread))
    {
      gdk_threads_leave();
      ochusha_async_buffer_wait(info->response_source_buffer,
				"bbs_thread_ui.c: each_response_cb()");
      gdk_threads_enter();
    }
#endif
  bbs_thread_view_append_text(view, buffer, -1);
  bbs_thread_view_append_text(view, LARGE_COLON, 3);
  bbs_thread_view_append_text(view, LARGE_COLON, 3);
  bbs_thread_view_append_text(view, _("[broken here]\n"), -1);

  bbs_thread_view_push_tag(view, context->response_paragraph_tag);
  bbs_thread_view_append_text(view, _("[broken here]\n"), -1);
  bbs_thread_view_pop_tags(view, context->response_paragraph_tag);

  gdk_threads_leave();

  return TRUE;
}


static gboolean
end_thread_rendering(OchushaBBSThread *thread, gboolean finished,
		     gpointer user_data)
{
  RendererContext *context = (RendererContext *)user_data;
  BBSThreadView *view = context->view;
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(context->thread);

  if (!finished)
    {
      /*
       * MEMO: å夬äƤ(ܡ󤵤줿쥹)
       *       å夵Ƥ̤ͤ꾯ʤäʤɡ
       *       DATեΤμľȤ롣
       */
#if DEBUG_GUI_MOST
      fprintf(stderr, "end_thread_rendering: clear current thread view\n");
#endif
      gdk_threads_enter();
      bbs_thread_view_clear(view);
      gdk_threads_leave();
    }
  else if (info->last_read == NULL)
    {
      /*
       * MEMO: ͭʤ
       *       äޤǤϿ쥹Ƭ˥ޡդƤΤǡ
       *       쥹ʤȥޡʤˤʤäƤޤΤǤ
       *       ХåեκǸ˥ޡĤƤ
       */
      info->last_read = bbs_thread_view_create_mark(view);
      info->next_mark = info->last_read;
      gdk_threads_enter();
      bbs_thread_view_scroll_to_mark(view, info->last_read);
      gdk_threads_leave();
#if DEBUG_GUI
      fprintf(stderr, "end_thread_rendering: What's happen?\n");
#endif
    }

#if DEBUG_GUI_MOST
  fprintf(stderr, "end_rendering\n");
#endif

  return TRUE;
}


static void
process_link(RendererContext *context)
{
  if (context->link == NULL)
    return;

  bbs_thread_view_append_text_as_link(context->view,
				      context->link_text_buffer,
				      context->link_text_length,
				      context->link);
  G_FREE(context->link);
  context->link = NULL;
}


static void
start_element_cb(void *user_data, const char *name, const char *const attrs[])
{
  RendererContext *context = (RendererContext *)user_data;
  BBSThreadView *view = context->view;
  int i;

#if DEBUG_GUI_MOST
  fprintf(stderr, "<%s", name);
  i = 0;
  while (attrs[i * 2] != NULL)
    {
      fprintf(stderr, " %s=%s", attrs[i * 2], attrs[i * 2 + 1]);
      i++;
    }
  fprintf(stderr, ">\n");
#endif
  bbs_thread_view_push_tag(view,
			   g_hash_table_lookup(context->tag_table, name));

  if (strcasecmp(name, "br") == 0)
    {
      bbs_thread_view_pop_tags(view, context->response_paragraph_tag);
      bbs_thread_view_append_text(view, "\n", 1);
      bbs_thread_view_push_tag(view, context->response_paragraph_tag);
    }
  else if (strcasecmp(name, "a") == 0)
    {
      gchar *link = NULL;
      if (context->link != NULL)
	{
#if DEBUG_GUI
	  fprintf(stderr, "previous link hasn't been closed.\n");
#endif
	  process_link(context);
	}
      context->link_text_buffer[0] = '\0';
      context->link_text_length = 0;

      i = 0;
      while (attrs[i * 2] != NULL)
	if (strcasecmp(attrs[i * 2], "href") == 0)
	  {
	    link = convert_string(context->handler.converter,
				  context->handler.helper,
				  attrs[i * 2 + 1], -1);
#if DEBUG_GUI_MOST
	    fprintf(stderr, "href=\"%s\"\n", link);
#endif
	    break;
	  }

      if (link == NULL)
	link = G_STRDUP(_("*empty link*"));

      context->link = link;
    }
}


static void
end_element_cb(void *user_data, const char *name)
{
  RendererContext *context = (RendererContext *)user_data;
  BBSThreadView *view = context->view;

#if DEBUG_GUI_MOST
  fprintf(stderr, "</%s>\n", name);
#endif
  bbs_thread_view_pop_tags(view,
			   g_hash_table_lookup(context->tag_table, name));

  if (strcasecmp(name, "a") == 0 && context->link != NULL)
    process_link(context);
}


static void
characters_cb(void *user_data, const char *ch, int len)
{
  gchar *buffer;
  int buffer_size;
  RendererContext *context = (RendererContext *)user_data;

  if (context->link == NULL)
    {
      bbs_thread_view_append_text(context->view, ch, len);
      return;
    }

  buffer = context->link_text_buffer;
  buffer_size = context->link_text_buffer_size;

  while (buffer_size <= (context->link_text_length + len))
    {
      fprintf(stderr, "buffer_size=%d, requirement=%d\n",
	      buffer_size, (context->link_text_length + len));
      buffer_size *= 2;
      if (context->link_text_buffer_size == DEFAULT_BUFFER_SIZE)
	{
	  buffer = (gchar *)G_MALLOC(buffer_size);
	  memcpy(buffer, context->link_text_buffer, context->link_text_length);
	}
      else
	buffer = (gchar *)G_REALLOC(buffer, buffer_size);

      context->link_text_buffer = buffer;
      context->link_text_buffer_size = buffer_size;
    }

  if (buffer_size > context->link_text_length + len)
    {
      memcpy(buffer + context->link_text_length, ch, len);
      context->link_text_length += len;
    }
  else
    {
      int left = buffer_size - context->link_text_length - 1;
      memcpy(buffer + context->link_text_length, ch, left);
      context->link_text_length = buffer_size - 1;
    }

  buffer[context->link_text_length] = '\0';
}


static gboolean
prepare_buffer_for_popup(OchushaApplication *application,
			 OchushaBBSThread *thread,
			 const gchar *link,
			 unsigned int *response_from,
			 unsigned int *response_to,
			 OchushaAsyncBuffer **buffer)
{
  unsigned int from = 0;
  unsigned int to = 0;
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(thread);

  if (application->maximum_number_of_popup_responses < 1)
    return FALSE;

  if (!ochusha_bbs_thread_check_url(thread, link, &from, &to))
    return FALSE;

#if DEBUG_GUI_MOST
  fprintf(stderr, "prepare_buffer_for_popup(): from=%d, to=%d\n", from, to);
#endif

  if ((from == 0 && to == 0) || from > thread->number_of_responses_read)
    return FALSE;

  if (to > thread->number_of_responses_read)
    to = thread->number_of_responses_read;

  if (to > 0 && from > to)
    {
      unsigned int tmp = to;
      to = from;
      from = tmp;
    }

  *response_from = from;
  *response_to = to;

  /*
   * ǶΥɤǤɽƤ륹ɬresponse_source_buffer˥Хåե
   * äƤ뤬⤷Ŭʥߥ󥰤ǲ٤⤷ʤΤǡ
   * Τ(å夫Τɤ߹)ɤĤƤ
   */
  if (info->response_source_buffer != NULL)
    {
      *buffer = info->response_source_buffer;
      return TRUE;
    }

  *buffer = ochusha_bbs_thread_get_responses_source(thread,
					application->broker,
					OCHUSHA_NETWORK_BROKER_CACHE_ONLY);
  info->response_source_buffer = *buffer;

  return (*buffer != NULL);
}


static gboolean need_popup_adjustment = FALSE;


static gboolean
popup_expose_event_cb(GtkWidget *widget, GdkEventExpose *event,
		      PopupContext *popup_context)
{
  GtkRequisition requisition = { 0, 0 };
  gtk_widget_size_request(widget, &requisition);


#if DEBUG_GUI_MOST
  fprintf(stderr, "popup_expose_event_cb(widget=%p), w=%d, h=%d\n",
	  widget, requisition.width, requisition.height);
#endif
  if (requisition.width > 0 && need_popup_adjustment)
    {
      BBS_THREAD_UI_LOCK
	{
	  need_popup_adjustment = FALSE;
	}
      BBS_THREAD_UI_UNLOCK;

      if (popup_context->renderer_context.application->enable_smart_popup_placement)
	{
	  GdkScreen *screen = gtk_widget_get_screen(popup_context->popup_window);

	  popup_context->y_pos -= requisition.height;
	  popup_context->y_pos
	    = popup_context->y_pos > 0 ? popup_context->y_pos : 0;

	  if (screen != NULL)
	    {
	      gint screen_width = gdk_screen_get_width(screen);
	      gint widget_width = requisition.width + THREAD_VIEW_EXTRA_WIDTH;
	      if (screen_width > 0
		  && (popup_context->x_pos + widget_width) > screen_width)
		{
		  popup_context->x_pos = screen_width - widget_width;
		  if (popup_context->x_pos < 0)
		    popup_context->x_pos = 0;
		}
	    }
	}

      gtk_widget_set_size_request(widget,
				  requisition.width
				  + THREAD_VIEW_EXTRA_WIDTH,
				  requisition.height);
      gtk_window_resize(GTK_WINDOW(popup_context->popup_window),
			requisition.width + THREAD_VIEW_EXTRA_WIDTH,
			requisition.height);

      if (popup_context->renderer_context.application->enable_smart_popup_placement)
	gtk_window_move(GTK_WINDOW(popup_context->popup_window),
			popup_context->x_pos, popup_context->y_pos);
    }

  return FALSE;
}


#if DEBUG_WIDGET_MOST
/*
 * XXX: Ȥʪˤʤ󥷥ʥġġ
 */
static gboolean
popup_map_event_cb(GtkWidget *widget, GdkEvent *event,
		   PopupContext *popup_context)
{
  GtkRequisition requisition = { 0, 0 };
  gtk_widget_size_request(widget, &requisition);
  fprintf(stderr, "popup_map_event_cb(widget=%p), w=%d, h=%d\n",
	  widget, requisition.width, requisition.height);
  return FALSE;
}


static void
popup_realize_cb(GtkWidget *widget, PopupContext *popup_context)
{
  GtkRequisition requisition = { 0, 0 };
  gtk_widget_size_request(widget, &requisition);
  fprintf(stderr, "popup_realize_cb(widget=%p), w=%d, h=%d\n",
	  widget, requisition.width, requisition.height);
  return;
}
#endif


static gboolean
popup_window_enter_notify_cb(GtkWidget *widget,
			     GdkEventCrossing *event,
			     PopupContext *popup_context)
{
#if DEBUG_POPUP
  fprintf(stderr, "popup_window_enter_notify_cb: widget=%p popup_context=%p\n",
	  widget, popup_context);
#endif
  if (popup_context->close_delay_id != 0)
    {
      g_source_remove(popup_context->close_delay_id);
      popup_context->close_delay_id = 0;
    }

  popup_context->now_pointed = TRUE;

  return FALSE;
}


static gboolean
popup_window_leave_notify_cb(GtkWidget *widget,
			     GdkEventCrossing *event,
			     PopupContext *popup_context)
{
#if DEBUG_POPUP
  fprintf(stderr, "popup_window_leave_notify_cb: widget=%p popup_context=%p\n",
	  widget, popup_context);
#endif
  popup_context->now_pointed = FALSE;

  hide_popup(popup_context);

  return FALSE;
}


static void
render_popup_response(PopupContext *popup_context,
		      OchushaBBSThread *thread, OchushaAsyncBuffer *buffer,
		      gint response_from, gint num_responses)
{
  OchushaApplication *application
    = popup_context->renderer_context.application;
  GtkWidget *view = bbs_thread_view_new(thread);
  bbs_thread_view_set_wrap_mode(BBS_THREAD_VIEW(view), GTK_WRAP_NONE);

  g_object_set_qdata(G_OBJECT(view), popup_parent_context_id,
		     popup_context);
#if DEBUG_POPUP
  fprintf(stderr, "render_popup_response(): view=%p\n", view);
#endif

  g_signal_connect(G_OBJECT(view), "link_mouse_over",
		   G_CALLBACK(link_mouse_over_cb), application);
  g_signal_connect(G_OBJECT(view), "link_mouse_out",
		   G_CALLBACK(link_mouse_out_cb), application);
  g_signal_connect(G_OBJECT(view), "link_mouse_press",
		   G_CALLBACK(link_mouse_press_cb), application);
  g_signal_connect(G_OBJECT(view), "link_mouse_release",
		   G_CALLBACK(link_mouse_release_cb), application);

  g_signal_connect(G_OBJECT(view), "view_enter_notify",
		   G_CALLBACK(popup_window_enter_notify_cb),
		   popup_context);
  g_signal_connect(G_OBJECT(view), "view_leave_notify",
		   G_CALLBACK(popup_window_leave_notify_cb),
		   popup_context);

  g_signal_connect_after(G_OBJECT(view), "expose_event",
			 G_CALLBACK(popup_expose_event_cb), popup_context);
#if DEBUG_WIDGET_MOST
  g_signal_connect(G_OBJECT(view), "map_event",
		   G_CALLBACK(popup_map_event_cb), popup_context);
  g_signal_connect_after(G_OBJECT(view), "realize",
			 G_CALLBACK(popup_realize_cb), popup_context);
#endif

  gtk_widget_show_all(view);

  if (popup_context->renderer_context.view != NULL)
    gtk_container_remove(popup_context->popup_frame,
			 GTK_WIDGET(popup_context->renderer_context.view));
  gtk_container_add(popup_context->popup_frame, view);
  gtk_widget_show(GTK_WIDGET(popup_context->popup_frame));

  popup_context->renderer_context.view = BBS_THREAD_VIEW(view);
  bbs_thread_view_hide_scrollbar(popup_context->renderer_context.view);
  popup_context->renderer_context.link_text_length = 0;
  popup_context->renderer_context.thread = thread;

#if DEBUG_GUI_MOST
  if (num_responses > 1)
    fprintf(stderr, "from=%d, num=%d\n", response_from, num_responses);
#endif

  popup_context->renderer_context.handler.converter
    = iconv_open("UTF-8//IGNORE",
		 ochusha_bbs_thread_get_response_character_encoding(thread));
  if (popup_context->renderer_context.handler.converter == (iconv_t)-1)
    {
      fprintf(stderr, "Out of memory: Couldn't render popup\n");
      return;
    }

  gdk_threads_leave();
  ochusha_bbs_thread_parse_responses(thread, buffer,
				     response_from - 1, num_responses,
				     NULL, render_response,
				     render_broken_response, NULL,
				     &popup_context->renderer_context);
  gdk_threads_enter();

  iconv_close(popup_context->renderer_context.handler.converter);
  popup_context->renderer_context.handler.converter = (iconv_t)0;
}


static void
show_popup(OchushaApplication *application, OchushaBBSThread *thread,
	   const gchar *link, gint x_pos, gint y_pos, BBSThreadView *parent)
{
  /*
   * MEMO: gdk_threads_enter()줿ĶƤФƤ롣
   */
  PopupContext *popup_context;
  OchushaAsyncBuffer *buffer;
  gint response_from = 0;
  gint response_to = 0;
  gint num_responses;

  popup_context = ensure_popup_context(parent, application);

#if DEBUG_GUI_MOST
  fprintf(stderr, "show_popup: link=\"%s\", x_pos=%d, y_pos=%d\n",
	  link, x_pos, y_pos);
#endif
  if (!prepare_buffer_for_popup(application, thread, link,
				&response_from, &response_to, &buffer))
    {
#if DEBUG_POPUP
      fprintf(stderr, "show_popup(): from=%d to=%d\n",
	      response_from, response_to);
      fprintf(stderr, "Popup cannot be prepared...why?\n");
#endif
      return;
    }

  if (response_from == 0)
    response_from
      = response_to - application->maximum_number_of_popup_responses;
  if (response_to == 0)
    response_to = response_from;

  num_responses = response_to - response_from + 1;
  if (num_responses > application->maximum_number_of_popup_responses)
    num_responses = application->maximum_number_of_popup_responses;

#if DEBUG_POPUP
      fprintf(stderr, "show_popup(): from=%d num_responses=%d\n",
	      response_from, num_responses);
#endif
  render_popup_response(popup_context, thread, buffer,
			response_from, num_responses);

  popup_context->x_pos = x_pos;
  popup_context->y_pos = y_pos;

  if (application->popup_response_delay == 0)
    show_popup_real(popup_context);
  else
    {
      if (popup_context->close_delay_id != 0)
	{
	  g_source_remove(popup_context->close_delay_id);
	  popup_context->close_delay_id = 0;
	}

      if (popup_context->delay_id != 0)
	g_source_remove(popup_context->delay_id);

      if (popup_context->url != NULL)
	G_FREE(popup_context->url);

      popup_context->url = G_STRDUP(link);
      popup_context->delay_id
	= g_timeout_add(application->popup_response_delay,
			(GSourceFunc)popup_delay_timeout, popup_context);
#if DEBUG_GUI_MOST
      fprintf(stderr, "timeout_add, id=%d, %p\n",
	      popup_context->delay_id, popup_context->url);
#endif
    }
}


static gboolean
popup_delay_timeout(PopupContext *popup_context)
{
  gboolean result = TRUE;

#if DEBUG_GUI_MOST
  fprintf(stderr, "popup_delay_timeout(%p)\n", data);
#endif

  gdk_threads_enter();

  if (popup_context->url != NULL)
    {
      if (popup_context->delay_id != 0)
	popup_context->delay_id = 0;

      G_FREE(popup_context->url);
      popup_context->url = NULL;

      show_popup_real(popup_context);
      result = FALSE;
    }

  gdk_threads_leave();

  return result;
}


static void
show_popup_real(PopupContext *popup_context)
{
#if DEBUG_POPUP
  fprintf(stderr, "show_popup_real(): popup_context=%p\n", popup_context);
#endif

  if (popup_context->renderer_context.view != NULL)
    {
      gint w, h;
      gboolean use_two_steps_hack = FALSE;
      GtkRequisition requisition = { 0, 0 };
      GdkScreen *screen = gtk_widget_get_screen(popup_context->popup_window);
      gint screen_width = screen != NULL ? gdk_screen_get_width(screen) : 0;

      gtk_widget_size_request(GTK_WIDGET(popup_context->renderer_context.view),
			      &requisition);
      gtk_widget_get_size_request(GTK_WIDGET(popup_context->renderer_context.view),
				  &w, &h);

      if (requisition.width == 0)
	{
	  BBS_THREAD_UI_LOCK
	    {
	      need_popup_adjustment = TRUE;
	    }
	  BBS_THREAD_UI_UNLOCK;

	  use_two_steps_hack = TRUE;
	}
      else
	{
	  gtk_widget_set_size_request(GTK_WIDGET(popup_context->renderer_context.view),
				      requisition.width
				      + THREAD_VIEW_EXTRA_WIDTH,
				      requisition.height);

	  gtk_window_resize(GTK_WINDOW(popup_context->popup_window),
			    requisition.width + THREAD_VIEW_EXTRA_WIDTH,
			    requisition.height);
	  BBS_THREAD_UI_LOCK
	    {
	      need_popup_adjustment = FALSE;
	    }
	  BBS_THREAD_UI_UNLOCK;
	}

      popup_context->y_pos -= (requisition.height + 10);
      popup_context->y_pos
	= popup_context->y_pos > 0 ? popup_context->y_pos : 0;

      if (popup_context->renderer_context.application->enable_smart_popup_placement
	  && screen_width > 0
	  && (popup_context->x_pos + requisition.width
	      + THREAD_VIEW_EXTRA_WIDTH) > screen_width)
	{
	  popup_context->x_pos =
	    screen_width - (requisition.width + THREAD_VIEW_EXTRA_WIDTH);
	  if (popup_context->x_pos < 0)
	    popup_context->x_pos = 0;
	}

      gtk_window_move(GTK_WINDOW(popup_context->popup_window),
		      popup_context->x_pos, popup_context->y_pos);
      if (use_two_steps_hack)
	popup_context->y_pos += requisition.height;

      gtk_widget_show_all(popup_context->popup_window);
#if DEBUG_POPUP
      fprintf(stderr, "show_popup_real(): show_all, view=%p\n",
	      popup_context->renderer_context.view);
#endif
    }
}


static void
hide_popup_real(PopupContext *popup_context)
{
  /*
   * MEMO: gdk_threads_enter()줿ĶƤФ롣
   */
  if (popup_context->url != NULL)
    {
      G_FREE(popup_context->url);
      popup_context->url = NULL;
    }

  if (popup_context->delay_id != 0)
    {
      g_source_remove(popup_context->delay_id);
      popup_context->delay_id = 0;
    }

#if DEBUG_POPUP
  fprintf(stderr, "hide_popup_real(): view=%p\n",
	  popup_context->renderer_context.view);
#endif
  gtk_widget_hide_all(popup_context->popup_window);
  if (popup_context->renderer_context.view != NULL)
    {
      gtk_container_remove(popup_context->popup_frame,
			   GTK_WIDGET(popup_context->renderer_context.view));
      popup_context->renderer_context.view = NULL;

      if (popup_context->parent_context != NULL)
	{
	  if (popup_context->parent_context != popup_context)
	    hide_popup(popup_context->parent_context);
	  else
	    fprintf(stderr, "hide_popup_real(): parent=%p current=%p\n",
		    popup_context->parent_context, popup_context);
	}
    }
}


static gboolean
popup_close_delay_timeout(PopupContext *popup_context)
{
  PopupContext *child_context = NULL;
#if DEBUG_POPUP
  fprintf(stderr, "popup_close_delay_timeout: popup_context=%p\n",
	  popup_context);
#endif

  gdk_threads_enter();

  if (!popup_context->now_pointed
      && !(popup_context->renderer_context.view != NULL
	   && (child_context = g_object_get_qdata(G_OBJECT(popup_context->renderer_context.view), popup_context_id)) != NULL
       && child_context->renderer_context.view != NULL))
    {
#if DEBUG_POPUP
      fprintf(stderr, "popup_close_delay_timeout: now_pointed=%d child=%p\n",
	      popup_context->now_pointed, child_context);
#endif
      hide_popup_real(popup_context);
    }

  popup_context->close_delay_id = 0;

  gdk_threads_leave();

  return FALSE;
}


static void
hide_popup(PopupContext *popup_context)
{
  /*
   * MEMO: gdk_threads_enter()ʴĶƤФ롣
   */
  OchushaApplication *application
    = popup_context->renderer_context.application;

#if DEBUG_POPUP
  fprintf(stderr, "hide_popup: popup_context=%p\n", popup_context);
#endif

  if (application->popup_close_delay == 0)
    hide_popup_real(popup_context);
  else if (popup_context->delay_id == 0)
    {
      if (popup_context->close_delay_id != 0)
	g_source_remove(popup_context->close_delay_id); /* Ԥ֥ꥻå */
      popup_context->close_delay_id
	= g_timeout_add(application->popup_close_delay,
			(GSourceFunc)popup_close_delay_timeout, popup_context);
    }
}


void
refresh_thread(OchushaApplication *application, GtkWidget *widget,
	       OchushaBBSThread *thread, IconLabel *tab_label)
{
  BBSThreadGUIInfo *info = ensure_bbs_thread_info(thread);
  BBSThreadView *view = BBS_THREAD_VIEW(widget);
  WorkerJob *job = NULL;
  BBSThreadJobArgs *job_args = NULL;
#if DEBUG_GUI_MOST
  fprintf(stderr, "refresh_thread\n");
#endif

  BBS_THREAD_UI_LOCK
    {
      OchushaAsyncBuffer *buffer;
      OchushaAsyncBuffer *old_buffer;

      if (!info->rendering_done)
	{
#if DEBUG_GUI_MOST
	  fprintf(stderr, "refresh_thread: refresh has already been requested.\n");
#endif
	  goto finish_refresh;	/* 󥰤򤷤Ƥ */
	}
      job = G_NEW0(WorkerJob, 1);
      job_args = G_NEW0(BBSThreadJobArgs, 1);
      old_buffer = info->response_source_buffer;
      buffer = ochusha_bbs_thread_get_responses_source(thread,
				application->broker,
				OCHUSHA_NETWORK_BROKER_CACHE_TRY_UPDATE);

      if (buffer == NULL)
	{
	  fprintf(stderr, "Couldn't access network!\n");
	  BBS_THREAD_UI_UNLOCK;
	  if (job != NULL)
	    G_FREE(job);
	  if (job_args != NULL)
	    G_FREE(job_args);
	  return;
	}

      setup_for_tab_label_animation(buffer, tab_label);

      info->response_source_buffer = buffer;
      info->rendering_done = FALSE;

      job_args->application = application;
      job_args->thread = thread;
      job_args->view = view;
      job_args->buffer = buffer;
      job_args->start = thread->number_of_responses_read;
      job_args->number_of_responses = -1;
      job_args->refresh = TRUE;
      job_args->tab_label = tab_label;

      job->canceled = FALSE;
      job->job = render_bbs_thread;
      job->args = job_args;

#if DEBUG_ASYNC_BUFFER_MOST
      fprintf(stderr, "refresh_thread: ref AsyncBuffer(%p)\n", buffer);
      fprintf(stderr, "* before unref: buffer->ref_count=%d\n",
	      G_OBJECT(buffer)->ref_count);
#endif
      g_object_ref(G_OBJECT(buffer));
      g_object_ref(G_OBJECT(tab_label));
      g_object_ref(G_OBJECT(view));
      commit_job(job);

      if (old_buffer != NULL)
	{
#if DEBUG_ASYNC_BUFFER_MOST
	  fprintf(stderr, "refresh_thread: unref AsyncBuffer(%p)\n",
		  old_buffer);
	  fprintf(stderr, "* before unref: old_buffer->ref_count=%d\n",
		  G_OBJECT(old_buffer)->ref_count);
#endif
	  g_object_unref(G_OBJECT(old_buffer));
	}
    }

 finish_refresh:
  BBS_THREAD_UI_UNLOCK;
  return;
}


void
go_to_the_first_response(GtkWidget *widget, OchushaBBSThread *thread)
{
  BBSThreadView *view;
  g_return_if_fail(widget != NULL);
  view = BBS_THREAD_VIEW(widget);
  bbs_thread_view_scroll_to_start(view);
}


void
go_to_the_last_response(GtkWidget *widget, OchushaBBSThread *thread)
{
  BBSThreadView *view;
  g_return_if_fail(widget != NULL);
  view = BBS_THREAD_VIEW(widget);
  bbs_thread_view_scroll_to_end(view);
}


void
jump_to_new_comer_response(GtkWidget *widget, OchushaBBSThread *thread)
{
  BBSThreadView *view;
  BBSThreadGUIInfo *info;
  g_return_if_fail(widget != NULL && OCHUSHA_IS_BBS_THREAD(thread));

  view = BBS_THREAD_VIEW(widget);
  info = ensure_bbs_thread_info(thread);
  if (info->last_read != NULL)
    bbs_thread_view_scroll_to_mark(view, info->last_read);
}
